import Stockholm from 'stockholm-js';
import parseNewick from '../parseNewick';
import { generateNodeIds } from '../util';
export default class StockholmMSA {
    data;
    MSA;
    constructor(text, currentAlignment) {
        const res = Stockholm.parseAll(text);
        this.data = res;
        this.MSA = res[currentAlignment];
    }
    getMSA() {
        return this.MSA;
    }
    getRow(name) {
        return this.MSA?.seqdata[name] || '';
    }
    getWidth() {
        const name = Object.keys(this.MSA?.seqdata)[0];
        return this.getRow(name).length;
    }
    get alignmentNames() {
        return this.data.map((aln, idx) => aln.gf.DE?.[0] || `Alignment ${idx + 1}`);
    }
    getHeader() {
        return {
            General: this.MSA.gf,
            Accessions: this.MSA.gs?.AC,
            Dbxref: this.MSA.gs?.DR,
        };
    }
    getRowData(rowName) {
        return {
            name: rowName,
            accession: this.MSA.gs?.AC?.[rowName],
            dbxref: this.MSA.gs?.DR?.[rowName],
        };
    }
    getNames() {
        return Object.keys(this.MSA.seqdata);
    }
    getSeqCoords() { }
    getStructures() {
        const pdbRegex = /PDB; +(\S+) +(\S); ([0-9]+)-([0-9]+)/;
        const ent = this.MSA;
        const args = Object.entries(ent.gs?.DR || {})
            .map(([id, dr]) => [id, pdbRegex.exec(dr)])
            .filter((item) => !!item[1])
            .map(([id, match]) => {
            const pdb = match[1].toLowerCase();
            const chain = match[2];
            const startPos = +match[3];
            const endPos = +match[4];
            return { id, pdb, chain, startPos, endPos };
        });
        const ret = {};
        for (const entry of args) {
            const { id, ...rest } = entry;
            if (!ret[id]) {
                ret[id] = [];
            }
            ret[id].push(rest);
        }
        return ret;
    }
    getTree() {
        const tree = this.MSA?.gf?.NH?.[0];
        return tree
            ? generateNodeIds(parseNewick(tree))
            : {
                id: 'root',
                name: 'root',
                noTree: true,
                branchset: this.getNames().map(name => ({
                    id: name,
                    branchset: [],
                    name,
                })),
            };
    }
    get seqConsensus() {
        return this.MSA.gc?.seq_cons;
    }
    get secondaryStructureConsensus() {
        return this.MSA.gc?.SS_cons;
    }
    get tracks() {
        return [
            {
                id: 'seqConsensus',
                name: 'Sequence consensus',
                data: this.seqConsensus,
                customColorScheme: {},
            },
            {
                id: 'secondaryStruct',
                name: 'Secondary-structure',
                data: this.secondaryStructureConsensus,
                customColorScheme: {
                    '>': 'pink',
                    '<': 'lightblue',
                },
            },
        ];
    }
}
//# sourceMappingURL=StockholmMSA.js.map