import React from 'react';
import { createRoot } from 'react-dom/client';
import { when } from 'mobx';
import { renderToStaticMarkup } from '@jbrowse/core/util';
import { renderTreeCanvas } from './components/tree/renderTreeCanvas';
import { renderMSABlock } from './components/msa/renderMSABlock';
import { colorContrast } from './util';
import MinimapSVG from './components/minimap/MinimapSVG';
import { renderBoxFeatureCanvasBlock } from './components/msa/renderBoxFeatureCanvasBlock';
export async function renderToSvg(model, opts) {
    await when(() => !!model.initialized);
    const { width, height, scrollX, scrollY } = model;
    const { exportType, theme, includeMinimap } = opts;
    if (exportType === 'entire') {
        return render({
            width: model.totalWidth + model.treeAreaWidth,
            height: model.totalHeight,
            theme,
            model,
            offsetY: 0,
            offsetX: 0,
            includeMinimap,
        });
    }
    else if (exportType === 'viewport') {
        return render({
            width,
            height,
            theme,
            model,
            offsetY: scrollY,
            offsetX: -scrollX,
            includeMinimap,
        });
    }
    else {
        throw new Error('unknown export type');
    }
}
async function render({ width, height, offsetX, offsetY, theme, model, includeMinimap, }) {
    const { Context } = await import('svgcanvas');
    const Wrapper = includeMinimap ? MinimapWrapper : NullWrapper;
    return renderToStaticMarkup(React.createElement(SvgWrapper, { width: width, height: height },
        React.createElement(Wrapper, { model: model },
            React.createElement(CoreRendering, { Context: Context, model: model, theme: theme, offsetX: offsetX, offsetY: offsetY, width: width, height: height }))), createRoot);
}
// function renderMultiline({ model }: { model: MsaViewModel }) {
//   const { treeAreaWidth, height } = model
//   const clipId = 'tree'
//   return (
//     <Wrapper model={model}>
//       <defs>
//         <clipPath id={clipId}>
//           <rect x={0} y={0} width={treeAreaWidth} height={height} />
//         </clipPath>
//       </defs>
//       <g
//         clipPath={`url(#${clipId})`}
//         /* eslint-disable-next-line react/no-danger */
//         dangerouslySetInnerHTML={{ __html: ctx1.getSvg().innerHTML }}
//       />
//       <g
//         transform={`translate(${treeAreaWidth} 0)`}
//         /* eslint-disable-next-line react/no-danger */
//         dangerouslySetInnerHTML={{ __html: ctx2.getSvg().innerHTML }}
//       />
//     </Wrapper>
//   )
// }
function CoreRendering({ model, theme, width, height, offsetX, offsetY, Context, }) {
    const clipId1 = 'tree';
    const clipId2 = 'msa';
    const { treeAreaWidth, colorScheme } = model;
    const contrastScheme = colorContrast(colorScheme, theme);
    const ctx1 = Context(width, height);
    const ctx2 = Context(width, height);
    renderBoxFeatureCanvasBlock({
        ctx: ctx2,
        offsetX,
        offsetY,
        model,
        blockSizeYOverride: height,
        highResScaleFactorOverride: 1,
    });
    const msaAreaWidth = width - treeAreaWidth;
    renderTreeCanvas({
        model,
        offsetY,
        ctx: ctx1,
        theme,
        blockSizeYOverride: height,
        highResScaleFactorOverride: 1,
    });
    renderMSABlock({
        model,
        theme,
        offsetY,
        offsetX,
        contrastScheme,
        ctx: ctx2,
        blockSizeXOverride: msaAreaWidth,
        blockSizeYOverride: height,
        highResScaleFactorOverride: 1,
    });
    return (React.createElement(React.Fragment, null,
        React.createElement("defs", null,
            React.createElement("clipPath", { id: clipId1 },
                React.createElement("rect", { x: 0, y: 0, width: treeAreaWidth, height: height }))),
        React.createElement("defs", null,
            React.createElement("clipPath", { id: clipId2 },
                React.createElement("rect", { x: 0, y: 0, width: msaAreaWidth, height: height }))),
        React.createElement("g", { clipPath: `url(#${clipId1})`, 
            /* eslint-disable-next-line react/no-danger */
            dangerouslySetInnerHTML: { __html: ctx1.getSvg().innerHTML } }),
        React.createElement("g", { clipPath: `url(#${clipId2})`, transform: `translate(${treeAreaWidth} 0)`, 
            /* eslint-disable-next-line react/no-danger */
            dangerouslySetInnerHTML: { __html: ctx2.getSvg().innerHTML } })));
}
function MinimapWrapper({ model, children, }) {
    const { minimapHeight, treeAreaWidth } = model;
    return (React.createElement(React.Fragment, null,
        React.createElement("g", { transform: `translate(${treeAreaWidth} 0)` },
            React.createElement(MinimapSVG, { model: model })),
        React.createElement("g", { transform: `translate(0 ${minimapHeight})` }, children)));
}
function SvgWrapper({ width, height, children, }) {
    return (React.createElement("svg", { width: width, height: height, xmlns: "http://www.w3.org/2000/svg", xmlnsXlink: "http://www.w3.org/1999/xlink", viewBox: [0, 0, width, height].toString() }, children));
}
function NullWrapper({ children }) {
    return children;
}
//# sourceMappingURL=renderToSvg.js.map